import gatt         # pip install dbus-python PyGObject gatt tqdm
from gatt.gatt_linux import Device
import time
import types
import re 
import tqdm
import argparse

t0 = 0
SERVICE_OTA_UUID    = '5833ff01-9b8b-5191-6142-22a4536ef123'
OTA_DATA_WRITE_UUID = "5833ff04-9b8b-5191-6142-22a4536ef123"
OTA_INDICATE_UUID   = "5833ff03-9b8b-5191-6142-22a4536ef123"
OTA_WRITE_UUID      = "5833ff02-9b8b-5191-6142-22a4536ef123"
mac = 'ab:cd:ef:ab:de:ff'
    
def checkSum(ib):
    # 注意这不是标准的CRC16    
    crc = 0
    for i2 in ib:
        # 0xb4 在java中取出来就是 -76 加上 0x100 后得到 180, python 下取到的直接就是 180 
        crc = crc^i2         
        for i3 in range(8):
            if crc&1!=0:
                crc = (crc>>1) ^ 0xa001     # 40961:  0xa001 
            else:
                crc = crc>>1    
    return crc 

def translateStr( s ): 
    # 交换字串顺序    
    bs = re.findall(r'\w{2}',s)
    bs.reverse()
    return ''.join(bs)
    
def int2str(v, bLong=True):
    return translateStr( '%08x' % v ) if bLong else  translateStr( '%04x' % v )

def loadHex(f):
    lines = [ l.strip()[1:] for l in open(f, 'r').readlines()]  # 去掉前面的 : 和结尾的 \r\n
    parts=[]
    for l in lines:
        # hex文件格式参考： https://developer.arm.com/documentation/ka003292/latest  
        ll,aaaa,tt,dd,crc  = int(l[:2], 16), l[2:6], l[6:8], l[8:-2], l[-2: ]    
        if   tt== '00': # data record
            parts[-1]['dat'].append( l[8:-2] )
        elif tt=='01':  # end-of-file record
            print('parse end')  
        elif tt=='04':  # extended linear address record
            parts.append( {'head': l, 'dat':[] } )                
        else: #  02: extended segment address record 05: start linear address record (MDK-ARM only)
            print( 'TODO', tt, l )
    
    dd0 = ''           
    for idx, part in enumerate(parts):
        dat = bytes.fromhex( ''.join( part['dat'] ) )        
        head = part['head']
        aaaa,tt, dd = head[2:6], head[6:8], head[8:-2]    
        partition_address = int(dd+'0000', 16 )     # partition_address
        flash_address = partition_address
        """
        scatter_load.sct 
        LR_ROM_XIP    0x11020000 0x020000   开始位置，最大空间              128K 看来这是放库的  
        LR_ROM_JT_GC  0x1fff0000 0x00800              最大空间只有 2K
        LR_IROM1      0x1fff1838 0x0E7C8    开始位置 0x1fff + 1838 ( 6200 ),  E7C8 不到58K
        """
        if dd!=dd0:               # 第一次
            dd0 = dd
            if dd=='1FFF':
                flash_address     = 0x0000
                partition_address = int(dd+'0000', 16 )                  
        else:       # 还处在同一个数据段, 那么地址就是 上一个的地址+ 上一个的长度
            part0 = parts[idx-1]  # 上一个part  
            if dd=='1FFF':        # 特别处理
                flash_address = part0['addr0'] + len(part0['dat']) + 8             
                partition_address = part0['addr1']+len(part0['dat']) 
                partition_end = partition_address + len(part['dat'])                
                if partition_end  > 0x1fff0800 and partition_address < 0x1fff0800:  # 要写入的段开始位置小于 1fff0800， 结束的位置却大于 1fff0800 时，需要修改 partition_address
                    partition_address = int('1fff1838', 16)                   
            else: # if (285212672 <= Long.parseLong(partition.getAddress(), 16) && Long.parseLong(partition.getAddress(), 16) <= 285736959){j = Long.parseLong(partition.getAddress(), 16); }
                partition_address = part0['addr1']+len(part0['dat'])   
                flash_address     = partition_address
        part['addr0'] = flash_address
        part['addr1'] = partition_address 
        part['dat']   = dat
        pre  = '02%02x' % idx  
        saddr0 = translateStr('%08x' % flash_address )
        saddr1 = translateStr(hex(partition_address)[2:])    
        slen = int2str( len( part['dat']), 8 )
        scrc = translateStr( hex( checkSum( dat ) )[2:] )
        part['cmd'] = ''.join([ pre , saddr0 , saddr1,  slen, scrc] )      
    return parts    
    
    
class AnyDeviceManager(gatt.DeviceManager):
    devices = []
    def device_discovered(self, device):
        if not device in self.devices:
            print('Discovered [%s] %s' % (device.mac_address, device.alias()) )
            self.devices.append( device )
        if time.time() - t0 > 5:
            self.stop()

class AnyDevice(gatt.Device):
    
    def connect_succeeded(self):
        super().connect_succeeded()
        print("[%s] Connected" % (self.mac_address))        # 1

    def connect_failed(self, error):
        super().connect_failed(error)
        print("[%s] Connection failed: %s" % (self.mac_address, str(error)))

    def disconnect_succeeded(self):
        super().disconnect_succeeded()
        print("[%s] Disconnected" % (self.mac_address))

    def services_resolved(self):
        super().services_resolved()
        print( 'services_resolved' )                        # 2 
        service = next(s for s in self.services if s.uuid==SERVICE_OTA_UUID) 
        characteristics = service.characteristics 
        characteristics.sort(key = lambda c:c.uuid) # 5833ff02， 5833ff03, 5833ff04
        characteristics[0].enable_notifications()   # CHARACTERISTIC_WRITE_UUID         sendOTACommand
        characteristics[1].enable_notifications()   # CHARACTERISTIC_OTA_INDICATE_UUID  收到特征值:
        for c in characteristics:
            print( c.uuid )
        self.characteristics = characteristics
        self.send()        
          
    def sendOTACommand(self, ob): # OTA_WRITE_UUID
        self.characteristics[0].write_value( ob )
    
    def sendOTAData(self, ob):    # OTA_DATA_WRITE_UUID
        self.characteristics[2].write_value( ob )     
    
    def characteristic_value_updated(self, characteristic, value):
        print( characteristic.uuid[:8], '< ', value.hex() )  

    def characteristic_write_value_failed(self, characteristic, error):
        print(characteristic.uuid[:8], ' Error', error )
        self.disconnect()

    def characteristic_write_value_succeeded(self, characteristic):
        self.send()

    def send(self):
        global tasks
        tp, ob = next(tasks, ('end',0) ) # 是没有数据时的默认值 ('end',0)
        if tp=='cmd':
            self.sendOTACommand( ob ) 
        elif tp=='bin':
            self.sendOTAData( ob )  
        else:
            print(tp, 'done') 
            self.disconnect()

if __name__ == '__main__':
    parser = argparse.ArgumentParser() 
    parser.add_argument('-file', default='ota/ec_open_mcu_app.hex', help='file of hex') 
    parser.add_argument('-mac',  default='ab:cd:ef:ab:de:ff', help='target mac') 
    parser.print_help()  
    args = parser.parse_args()
    path = args.file
    mac = args.mac
    
    parts = loadHex(path)
    tasks = []
    tasks.append( ('cmd', bytes.fromhex('010700')) ) 
    for part in parts:
        tasks.append( ('cmd', bytes.fromhex( part['cmd'])) )  # <  b'\x00\x84'    
        dat =  part['dat']      
        for i in range(0, len( dat ), 237):         # 237 字节 69*237+31 = 16384 = 16K
            tasks.append( ('bin', dat[i:i+237]) ) 
    tasks.append( ('cmd', bytes.fromhex('0401')) )
    tasks = iter(tasks)   
    manager = gatt.DeviceManager(adapter_name='hci0')
    device = AnyDevice(mac_address=mac, manager=manager)
    device.connect()
    manager.run()


"""
参考反编译内容: 

cmd 010700                              0081 
cmd 02000000021100000211004000005abe    0084 
dat LINE--237B                          0084  为什么我的没有 用的 characteristic 不对吗?
    行17
    又有很多行 数据发送成功： 0087
dat 半行数据                           0085    之后发下一个 part 的头部命令
                                        0087  
                                        0083   otaUtilsCallBack.onOTAFinish() or otaUtilsCallBack.onResourceFinish()
cmd 0401                                008A   otaUtilsCallBack.onReBootSuccess()    断开连接  


    写 OTA_WRITE_UUID 成功后会收到：
    0081：
        sendPartition(bluetoothGatt, this.firmWareFile, this.partitionIndex, this.mFlash_addr); 
        this.blockIndex = 0;                                                                        or
        sendResource(bluetoothGatt, this.firmWareFile);
    0084:  发送第一个数据专
        List<String> list = this.firmWareFile.getList().get(this.partitionIndex).getBlocks().get(this.blockIndex);
        sendOTAData(bluetoothGatt, list.get( 0 )); 
    0089：                  
        sendPartition(bluetoothGatt, this.firmWareFile, this.partitionIndex, this.mFlash_addr);
        this.blockIndex = 0;
    0102 or 0103
        Log.d(str, "start ota or resource");            
    以下是34个字节长的响应    
    71xxxxxxxx...     
        BleUtils.sendOTACommand(bluetoothGatt, "06" + this.randomStr, true);
    72xxxxxxx
        String decrypt = AESTool.decrypt(this.firwareData, this.password);
        BleUtils.sendOTACommand(bluetoothGatt, "07" + AESTool.encrypt(AESTool.encrypt(decrypt, this.randomStr), this.password), true);
    73xxxxxx 
        BleUtils.sendOTACommand(bluetoothGatt, "0102", true);
    8Bxxxxxx
        this.firwareData = this.response.substring(2);
        BleUtils.sendOTACommand(bluetoothGatt, "07" + this.randomStr, true);
    8Cxxxxxx
        String decrypt2 = AESTool.decrypt(this.firwareData, this.password);
        BleUtils.sendOTACommand(bluetoothGatt, "08" + AESTool.encrypt(AESTool.encrypt(decrypt2, this.randomStr), this.password), true);
    8Dxxxxxx
        this.otaUtilsCallBack.onStartSecurityData();
            
    写 OTA_DATA_WRITE_UUID 成功后 
        if(i==0){
            float length = this.finshSize + bluetoothGattCharacteristic.getValue().length;
            this.otaUtilsCallBack.onProcess((length * 100.0f) / this.totalSize);
            
            this.cmdIndex = this.cmdIndex + 1;
            if (this.cmdIndex < this.cmdList.size()) 
                sendOTAData(bluetoothGatt, this.cmdList.get(this.cmdIndex));
        }else{
            retryCmd(bluetoothGatt, 1004);--> sendOTAData(bluetoothGatt, this.cmdList.get(this.cmdIndex));
        }


    public void onCharacteristicChanged(BluetoothGatt bluetoothGatt, BluetoothGattCharacteristic bluetoothGattCharacteristic) 
    0087:
        this.blockIndex +=1;
        this.cmdIndex = 0;
        List<String> list = this.firmWareFile.getList().get(this.partitionIndex).getBlocks().get(this.blockIndex);
        sendOTAData(bluetoothGatt, list.get(this.cmdIndex));
    0085:
                int i2 = this.partitionIndex + 1;
                this.partitionIndex = i2;
                this.blockIndex = 0;
                if (i2 < this.firmWareFile.getList().size()) {
                    if (this.otaType == OTAType.OTA || this.otaType == OTAType.Security) {
                        Partition partition = this.firmWareFile.getList().get(this.partitionIndex - 1);
                        if (285212672 > Long.parseLong(partition.getAddress(), 16) || Long.parseLong(partition.getAddress(), 16) > 285736959) {
                            if (this.firmWareFile.getPath().endsWith("hexe16")) {
                                this.mFlash_addr = this.mFlash_addr + partition.getPartitionLength() + 4;
                            } else {
                                this.mFlash_addr = this.mFlash_addr + partition.getPartitionLength() + 8;
                            }
                        }
                    }
                    sendPartition(bluetoothGatt, this.firmWareFile, this.partitionIndex, this.mFlash_addr);
                }
            }
                                     
"""



            

